#define _POSIX_C_SOURCE 200809L

#include <lektor/database.h>
#include <lektor/macro.h>
#include <stdio.h>
#include <strings.h>

/* Find it in database/open.c */
extern int
is_sql_str_invalid(const char *str);

bool
database_plt_create(sqlite3 *db, const char *name)
{
    static const char *SQL_STMT =
        "INSERT INTO playlist (name, last_update) VALUES (?, strftime('%s', 'now'));";
    sqlite3_stmt *stmt = NULL;
    bool sta = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);

    if (sqlite3_step(stmt) != SQLITE_DONE) {
        fprintf(stderr, " ! database_plt_create: expected SQLITE_DONE in step: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    sta = true;
    fprintf(stderr, " . database_plt_create: Created playlist %s\n", name);
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_remove(sqlite3 *db, const char *name)
{
    static const char *SQL_STMT =
        "DELETE FROM playlist WHERE name = ?;";
    sqlite3_stmt *stmt = NULL;
    bool sta = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);

    if (sqlite3_step(stmt) != SQLITE_DONE) {
        fprintf(stderr, " ! database_plt_remove: expected SQLITE_DONE in step: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_remove_pos(sqlite3 *db, const char *name, int pos)
{
    static const char *SQL_STMT =
        "DELETE FROM kara_playlist WHERE playlist_id = (SELECT id FROM playlist WHERE name = ? LIMIT 1) and kara_id = ?;";
    sqlite3_stmt *stmt = NULL;
    bool sta = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_BIND_INT(db, stmt, 2, pos, error);

    if (sqlite3_step(stmt) != SQLITE_OK) {
        fprintf(stderr, " ! database_plt_remove_pos: expected SQLITE_DONE in step: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_clear(sqlite3 *db, const char *name)
{
    static const char *SQL_STMT =
        "DELETE FROM kara_playlist WHERE playlist_id = (SELECT id FROM playlist WHERE name = ?);";
    sqlite3_stmt *stmt = NULL;
    bool sta = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);

    if (sqlite3_step(stmt) != SQLITE_OK) {
        fprintf(stderr, " ! database_plt_clear: expected SQLITE_DONE in step: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_rename(sqlite3 *db, const char *old_name, const char *new_name)
{
    static const char *SQL_STMT =
        "UPDATE playlist SET name = ? WHERE name = ?;";
    sqlite3_stmt *stmt = NULL;
    bool sta = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, new_name, error);
    SQLITE_BIND_TEXT(db, stmt, 2, old_name, error);

    if (sqlite3_step(stmt) != SQLITE_OK) {
        fprintf(stderr, " ! database_plt_rename: expected SQLITE_DONE in step: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_export(sqlite3 *db, const char *name)
{
    static const char *SQL_SCHEM =
        "CREATE TABLE IF NOT EXISTS %s.content"
        "( kara_id INTEGER PRIMARY KEY NOT NULL CHECK(kara_id > 0) );";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    int code, ret = false;
    sqlite3_stmt *stmt;

    if (is_sql_str_invalid(name)) {
        fprintf(stderr, " . database_plt_export: Name of playlist '%s' is invalid\n",
                name);
        goto error;
    }

    if (!strcasecmp(name, PROTECTED_DATABASE)) {
        fprintf(stderr, " . database_plt_export: Name '" PROTECTED_DATABASE
                "' is protected, you can't use it for a playlist name\n");
        goto error;
    }

    snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT - 1, SQL_SCHEM, name);
    SQL_STMT[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    code = sqlite3_step(stmt);

    if (code != SQLITE_DONE && code != SQLITE_OK) {
        fprintf(stderr, " . database_plt_export: Failed to create schema: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_plt_import(sqlite3 *db, const char *name)
{
    static const char *SQL_SCHEM =
        "INSERT OR IGNORE INTO kara_playlist (kara_id, playlist_id)"
        " SELECT %s.content.kara_id, ( SELECT id FROM playlist WHERE name = %s LIMIT 1 )"
        " FROM %s.content;\n";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT], ret = false;

    if (is_sql_str_invalid(name)) {
        fprintf(stderr, " . database_plt_import: Name of playlist '%s' is invalid\n",
                name);
        goto error;
    }

    if (!strcasecmp(name, PROTECTED_DATABASE)) {
        fprintf(stderr, " . database_plt_import: Name '" PROTECTED_DATABASE
                "' is protected, you can't use it for a playlist name\n");
        goto error;
    }

    snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT - 1, SQL_SCHEM, name, name, name);
    SQL_STMT[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_EXEC(db, SQL_STMT, error);
    ret = true;
error:
    return ret;
}

bool
database_plt_add_uri(sqlite3 *db, const char *name, struct lkt_uri_t *uri)
{
    static const char *SQL =
        "WITH plt_id(id) AS (SELECT playlist.id AS id FROM playlist WHERE name COLLATE nocase = ?) "
        "INSERT INTO kara_playlist (kara_id, playlist_id) "
        "SELECT plt_id.id, kara.id "
        "  FROM kara, plt_id "
        "  WHERE kara.%s LIKE ? "
        "  ORDER BY RANDOM();";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT], sta = false;
    const char *column;
    sqlite3_stmt *stmt;

    switch (uri->type) {
    case uri_type:
        column = "category";
        break;
    case uri_author:
        column = "author_name";
        break;
    case uri_category:
        column = "type";
        break;
    case uri_language:
        column = "language";
        break;
    case uri_query:
    case uri_fs:
        column = "file_path";
        break;
    default:
        return false;
    }

    snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT - 1, SQL, column);
    SQL_STMT[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_BIND_TEXT(db, stmt, 2, (char *) uri->value, error);

    if (sqlite3_step(stmt) != SQLITE_DONE) {
        fprintf(stderr, " ! database_plt_add_uri: expected SQLITE_DONE in step: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}
