#define _POSIX_C_SOURCE 200809L
#define _DEFAULT_SOURCE

#include <lektor/database.h>
#include <lektor/macro.h>
#include <stdbool.h>
#include <sqlite3.h>
#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <linux/limits.h>

static bool
database_add_kara(sqlite3 *db, const char *filename)
{
    if (db == NULL || filename == NULL) {
        fprintf(stderr, " ! database_add_kara: Invalid arguments\n");
        return false;
    }

    static const char *SQL_STMT =
        "INSERT INTO "
        "kara (song_name, source_name, category, song_type, language, file_path, is_new, author_name, author_year, song_number)"
        "SELECT ?, ?, ?, ?, ?, ?, ?, ?, strftime('%s','now'), ?";
    sqlite3_stmt *stmt = NULL;
    bool status = false;
    int code = SQLITE_OK;
    struct kara_metadata data;
    time_t the_time = time(NULL);
    struct tm *the_local_time = localtime(&the_time);
    char year[10];
    code = snprintf(year, 10, "%d", the_local_time->tm_year + 1900);

    if (code < 0 || code >= 10) {
        fprintf(stderr, " ! database_add_kara: Failed to get the year of the current date: %s\n", sqlite3_errmsg(db));
        return false;
    }

    if (kara_metadata_read(&data, filename) != 0) {
        fprintf(stderr, " ! database_add_kara: Failed to get matadata of the kara %s: %s\n", filename,
                sqlite3_errmsg(db));
        return false;
    }

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);

    if ((sqlite3_bind_text(stmt, 1, data.song_name, -1, 0) != SQLITE_OK)    ||
        (sqlite3_bind_text(stmt, 2, data.source_name, -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 3, data.category, -1, 0) != SQLITE_OK)     ||
        (sqlite3_bind_text(stmt, 4, data.song_type, -1, 0) != SQLITE_OK)    ||
        (sqlite3_bind_text(stmt, 5, data.language, -1, 0) != SQLITE_OK)     ||
        (sqlite3_bind_text(stmt, 6, filename, -1, 0) != SQLITE_OK)          ||
        (sqlite3_bind_int(stmt, 7, 0) != SQLITE_OK)                         ||
        (sqlite3_bind_text(stmt, 8, data.author_name, -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_int(stmt,  9, data.song_number) != SQLITE_OK)
       ) {
        fprintf(stderr, " ! database_add_kara: failed to bind argument to stmt statement for kara %s: %s\n",
                filename, sqlite3_errmsg(db));
        goto error;
    }

    if (sqlite3_step(stmt) != SQLITE_DONE) {
        fprintf(stderr,
                " ! database_add_kara: expected SQLITE_DONE after execution of stmt statement for kara %s: %s\n",
                filename,
                sqlite3_errmsg(db));
        goto error;
    }

    status = true;
error:
    sqlite3_finalize(stmt);
    return status;
}

bool
database_update_add(sqlite3 *db, const char *kara_path, struct kara_metadata *mdt, uint64_t id, bool avail)
{
    if (db == NULL || kara_path == NULL || mdt == NULL) {
        fprintf(stderr, " ! database_add_kara: Invalid arguments\n");
        errno = EINVAL;
        return false;
    }

    static const char *SQL_STMT =
        "INSERT OR REPLACE INTO "
        "kara (song_name, source_name, category, song_type, language, file_path, is_new, author_name, author_year, song_number, id, available)"
        "SELECT ?, ?, ?, ?, ?, ?, ?, ?, strftime('%s','now'), ?, ?, ?";
    sqlite3_stmt *stmt = NULL;
    int code = SQLITE_OK;
    bool ret = false;
    time_t the_time = time(NULL);
    struct tm *the_local_time = localtime(&the_time);
    char year[10];
    code = snprintf(year, 10, "%d", the_local_time->tm_year + 1900);

    if (code < 0 || code >= 10) {
        fprintf(stderr, " ! database_update_add: Failed to get the year of the current date: %s\n",
                sqlite3_errmsg(db));
        goto error_no_sqlite;
    }

    /* From here we initialize the sqlite stmt. */

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);

    if ((sqlite3_bind_text(stmt, 1, mdt->song_name, -1, 0) != SQLITE_OK)    ||
        (sqlite3_bind_text(stmt, 2, mdt->source_name, -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 3, mdt->category, -1, 0) != SQLITE_OK)     ||
        (sqlite3_bind_text(stmt, 4, mdt->song_type, -1, 0) != SQLITE_OK)    ||
        (sqlite3_bind_text(stmt, 5, mdt->language, -1, 0) != SQLITE_OK)     ||
        (sqlite3_bind_text(stmt, 6, kara_path, -1, 0) != SQLITE_OK)         ||
        (sqlite3_bind_int(stmt, 7, 0) != SQLITE_OK) /* TODO */              ||  /* No new kara added (TODO). */
        (sqlite3_bind_text(stmt, 8, mdt->author_name, -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_int(stmt,  9, mdt->song_number) != SQLITE_OK)         ||
        (sqlite3_bind_int(stmt, 10, id) != SQLITE_OK)                       ||
        (sqlite3_bind_int(stmt, 11, avail) != SQLITE_OK)
       ) {
        fprintf(stderr, " ! database_update_add: failed to bind argument to stmt statement for kara %s: %s\n",
                kara_path, sqlite3_errmsg(db));
        goto error;
    }

    if (sqlite3_step(stmt) != SQLITE_DONE) {
        fprintf(stderr, " ! database_add_kara: expected sqlite_done after execution of stmt "
                "statement for kara %s: %s\n", kara_path, sqlite3_errmsg(db));
        goto error;
    }

    ret = true;
error:
    sqlite3_finalize(stmt);
error_no_sqlite:
    return ret;
}

bool
database_update(sqlite3 *db, const char *kara_dir)
{
    DIR *d;
    struct dirent *dir;
    char path[PATH_MAX];

    memset(path, 0, PATH_MAX * sizeof(char));
    d = opendir(kara_dir);

    if (!d) {
        fprintf(stderr, " ! database_update: Failed to open directory '%s': %s\n", kara_dir, strerror(errno));
        return false;
    }

    while (NULL != (dir = readdir(d))) {
        strncpy(path, kara_dir, PATH_MAX - 1);
        strncat(path, "/", PATH_MAX - 1);
        strncat(path, dir->d_name, PATH_MAX - 1);

        if (dir->d_type == DT_REG)
            database_add_kara(db, path);
        else if (dir->d_type == DT_DIR &&
                 strcmp(dir->d_name, ".") != 0 &&
                 strcmp(dir->d_name, "..") != 0)
            database_update(db, path);
    }

    fprintf(stderr, " * database_update: Passed directory '%s'\n", kara_dir);
    closedir(d);
    return true;
}
