#include <lektor/module/module_x11.h>
#include <lektor/database.h>
#include <lektor/commands.h>

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <mpv/client.h>

#define WIDTH   400
#define HEIGHT  200

struct module_x11_window {
    /* X11 related, the master and slave window */
    Display *master_display;
    Display *child_display;
    Window master_win;
    Window child_win;

    /* X11 related, some X11 properties */
    int screen;
    XSetWindowAttributes attrs;
    Atom wm_delete;
    XColor black;
    bool is_fullscreen;

    /* Mpv related */
    mpv_handle *mpv;
    volatile int mpv_time_pos;  // Don't write it in the database       //
    volatile int mpv_duration;  // Because don't need to be persistent  //
};

/*
 * NOW: THE ONLY EXPORTED FUNCTION :NOW
 */

extern int
module_set_function(void *arg__, void *handle__)
{
    if (NULL == arg__ || handle__ == NULL)
        return 1;

    struct lkt_win *win = (struct lkt_win *) arg__;

    win->new = module_x11_new;
    win->close = module_x11_close;
    win->free = module_x11_free;
    win->toggle_pause = module_x11_toggle_pause;
    win->load_file = module_x11_load_file;
    win->set_volume = module_x11_set_volume;
    win->is_paused = module_x11_is_paused;
    win->get_duration = module_x11_get_duration;
    win->get_elapsed = module_x11_get_elapsed;
    win->handle_events = module_x11_handle_events;
    win->handle = handle__;

    return 0;
}

/*
 *
 * NOW: PRIVATE X11 FUNCTIONS :NOW
 *
 */

static inline bool
lx11_new(struct module_x11_window *ret)
{
    ret->master_display = XOpenDisplay(NULL);
    if (ret->master_display == NULL) {
        fprintf(stderr, " ! lkt_x11: can't open display\n");
        return false;
    }

    ret->screen = DefaultScreen(ret->master_display);
    Colormap colormap = DefaultColormap(ret->master_display, ret->screen);
    XAllocNamedColor(ret->master_display, colormap, "black", &ret->black, &ret->black);

    // Initialize container window attributes
    ret->attrs.event_mask
        = SubstructureRedirectMask                  // handle child window requests         (MapRequest)
          | SubstructureNotifyMask                  // handle child window notifications    (DestroyNotify)
          | StructureNotifyMask                     // handle container notifications       (ConfigureNotify)
          | ExposureMask                            // handle container redraw              (Expose)
          | KeyReleaseMask                          // handle release key                   (KeyRelease)
          | VisibilityChangeMask                    // handle... I don't know
          ;
    ret->attrs.do_not_propagate_mask = 1;           // do not hide any events from child window
    ret->attrs.background_pixel = ret->black.pixel; // background color
    unsigned long attrs_mask = CWEventMask          // enable attrs.event_mask
                               | NoEventMask        // enable attrs.do_not_propagate_mask
                               | CWBackPixel        // enable attrs.background_pixel
                               ;

    // Create and map container window
    ret->master_win = XCreateWindow(ret->master_display, RootWindow(ret->master_display, ret->screen),
                                    0, 0, WIDTH, HEIGHT, 1, CopyFromParent, InputOutput,
                                    CopyFromParent, attrs_mask, &ret->attrs);

    XMapWindow(ret->master_display, ret->master_win);                               // Make window visible
    XStoreName(ret->master_display, ret->master_win, "Lektor Player");              // Set window title
    ret->wm_delete = XInternAtom(ret->master_display, "WM_DELETE_WINDOW", True);    // Get WM_DELETE_WINDOW atom
    XSetWMProtocols(ret->master_display, ret->master_win, &ret->wm_delete, 1);

    ret->child_display = NULL;
    ret->child_win = 0;

    return true;
}

static inline void
lx11_fullscreen(struct module_x11_window *win, bool go_fullscreen)
{
    if (go_fullscreen) {
        XSizeHints *size_hints = XAllocSizeHints();
        long hints = 0;

        if (NULL == size_hints) {
            fprintf(stderr, "lx11_fullscreen: failed to allocate size hints\n");
            return;
        }

        if (XGetWMSizeHints(win->master_display, win->master_win, size_hints, &hints,
                            XInternAtom(win->master_display, "WM_SIZE_HINTS", False))) {
            fprintf(stderr, " ! lx11_fullscreen: failed to get size hints\n");
            goto error;
        }

        XLowerWindow(win->master_display, win->master_win);
        XUnmapWindow(win->master_display, win->master_win);
        XSync(win->master_display, False);

        fprintf(stderr, " . lx11_fullscreen: size hints is: %ld\n", hints);
error:
        XFree(size_hints);
    }

    Atom atoms[2] = { XInternAtom(win->master_display, "_NET_WM_STATE_FULLSCREEN", go_fullscreen), None };
    XChangeProperty(win->master_display, win->master_win,
                    XInternAtom(win->master_display, "_NET_WM_STATE", False),
                    XA_ATOM, 32, PropModeReplace, (const unsigned char *) atoms, 1);
    XMapWindow(win->master_display, win->master_win);
    XRaiseWindow(win->master_display, win->master_win);
}

static inline bool
lx11_handle(struct module_x11_window *win)
{
    XEvent event;

    if (!win || !win->mpv)
        return true;

    for (;;) {

        if (!XCheckWindowEvent(win->master_display, win->master_win, win->attrs.event_mask, &event))
            return true;

        // Map child window when it requests and store its display and window id
        if (event.type == MapRequest) {
            XMapWindow(event.xmaprequest.display, event.xmaprequest.window);

            win->child_display = event.xmaprequest.display;
            win->child_win = event.xmaprequest.window;
        }

        // Propagate resize event to child window, and also resize it after MapRequest
        if (event.type == ConfigureNotify || event.type == MapRequest) {
            if (win->child_win) {
                // Get container window attributes
                XWindowAttributes attrs;
                XGetWindowAttributes(win->master_display, win->master_win, &attrs);

                // Move and resize child
                XMoveResizeWindow(win->child_display, win->child_win,
                                  2, 2, attrs.width - 6, attrs.height - 6);
            }
        }

        // Key press
        if (event.type == KeyRelease) {
            fprintf(stderr, " . lx11_handle: released key with keycode: 0x%x\n", event.xkey.keycode);
            // if (event.xkey.keycode == 0x5f)
            //     lx11_fullscreen(win, true);
            //
            if (event.xkey.keycode == 0x47)
                XClearWindow(win->master_display, win->master_win);
        }

        // Refresh container window
        if (event.type == Expose)
            XClearWindow(win->master_display, win->master_win);

        // Close button
        if (event.type == ClientMessage) {
            if ((long unsigned int) event.xclient.data.l[0] == win->wm_delete)
                return false;
        }

    }
    return true;
}

/*
 *
 * NOW: PRIVATE MPV FUNCTIONS :NOW
 *
 */

static inline void
lmpv_free(mpv_handle **ctx)
{
    if (!*ctx) {
        fprintf(stderr, " ! lmpv_free: failed due to missing mpv ctx\n");
        return;
    }

    static const char *cmd[] = {"quit", NULL};
    mpv_command(*ctx, cmd);
    mpv_destroy(*ctx);
    *ctx = NULL;
    fprintf(stderr, " * lmpv_free: mpv context destroyed\n");
}

static inline bool
lmpv_handle(struct lkt_win *win, sqlite3 *db, enum mpd_idle_flag *mpd_idle_events)
{
    int ao_volume;
    struct lkt_queue_state state;
    bool sta = false;
    struct module_x11_window *win_x11 = win->window;
    mpv_event *event = NULL;
    mpv_event_property *prop;

    if (!win || !win_x11 || !win_x11->mpv)
        goto end;

    if (!database_queue_state(db, &state))
        goto end;

    for (;;) {
        event = mpv_wait_event(win_x11->mpv, 0);

        switch (event->event_id) {
        case MPV_EVENT_PAUSE:
        case MPV_EVENT_UNPAUSE:
            database_queue_toggle_pause(db);
            break;

        case MPV_EVENT_SHUTDOWN:
            database_queue_stop(db);
            win->close(win);
            goto end;

        case MPV_EVENT_NONE:
            goto end;

        case MPV_EVENT_IDLE:
            if (state.current > 0 && win_x11->mpv_time_pos > 0)
                command_next(db, win, mpd_idle_events);
            break;

        case MPV_EVENT_PROPERTY_CHANGE:
            prop = (mpv_event_property *) event->data;
            if (prop->format == MPV_FORMAT_NONE)
                break;
            // MPV volume (BUG: The flag is not MPV_FORMAT_NONE only at the end of the song...) //
            if (!strcmp(prop->name, "ao-volume")
                && prop->format == MPV_FORMAT_INT64) {
                ao_volume = *(int *) prop->data;
                database_config_queue(db, "volume", ao_volume);
            }
            // File duration //
            if (!strcmp(prop->name, "duration")
                && prop->format == MPV_FORMAT_INT64) {
                win_x11->mpv_duration = *(int *) prop->data;
                database_config_queue(db, "duration", *(int *) prop->data);
            }
            if (!strcmp(prop->name, "time-pos")
                && prop->format == MPV_FORMAT_INT64) {
                win_x11->mpv_time_pos = *(int *) prop->data;
                database_config_queue(db, "elapsed", *(int *) prop->data);
            }
            // Pause state //
            if (!strcmp(prop->name, "pause")
                && prop->format == MPV_FORMAT_FLAG)
                database_queue_set_paused(db, *(bool *) prop->data);
            break;
        default:
            fprintf(stderr, " ! Unhandled mpv event: %s\n", mpv_event_name(event->event_id));
            break;
        }
    }

    sta = true;
end:
    return sta;
}

static mpv_handle *
lmpv_new(unsigned long int wid)
{
    mpv_handle *ctx = mpv_create();
    int status;

    if (!ctx) {
        fprintf(stderr, " ! lmpv_new: failed creating context\n");
        return NULL;
    }

    if ((status = mpv_set_option_string(ctx, "input-default-bindings", "yes")) < 0) {
        fprintf(stderr, " ! lmpv_new: failed to set input-default-bindings: %s\n",
                mpv_error_string(status));
        return NULL;
    }

    if ((status = mpv_set_option_string(ctx, "input-vo-keyboard", "yes")) < 0) {
        fprintf(stderr, " ! lmpv_new: failed to set input-vo-keyboard: %s\n",
                mpv_error_string(status));
        return NULL;
    }

    if ((status = mpv_set_option_string(ctx, "replaygain", "track")) < 0) {
        fprintf(stderr, " ! lmpv_new: failed to set replaygain: %s\n",
                mpv_error_string(status));
        return NULL;
    }

    if ((status = mpv_set_option_string(ctx, "gpu-context", "x11")) < 0) {
        fprintf(stderr, " ! lmpv_new: failed to set gpu-context: %s\n",
                mpv_error_string(status));
        return NULL;
    }

    if ((status = mpv_set_property(ctx, "wid", MPV_FORMAT_INT64, &wid)) < 0) {
        fprintf(stderr, " ! lmpv_new: failed to set wid: %s\n",
                mpv_error_string(status));
        return NULL;
    }

    if ((status = mpv_set_option_string(ctx, "demuxer-max-bytes", "10M")) < 0) {
        fprintf(stderr, " ! lmpv_new: failed to set demuxer-max-bytes: %s\n",
                mpv_error_string(status));
        return NULL;
    }

    if ((status = mpv_set_option_string(ctx, "profile", "low-latency")) < 0) {
        fprintf(stderr, " ! lmpv_new: failed to set demuxer-max-bytes: %s\n",
                mpv_error_string(status));
        return NULL;
    }

    if ((status = mpv_set_property_string(ctx, "hwdec", "yes")) < 0) {
        fprintf(stderr, " ! lmpv_new: failed to set hwdec to no: %s\n",
                mpv_error_string(status));
        return NULL;
    }

    if ((status = mpv_initialize(ctx)) < 0) {
        fprintf(stderr, " ! lmpv_new: failed to initialize mpv: %s\n",
                mpv_error_string(status));
        return NULL;
    }

    if ((status = mpv_observe_property(ctx, 0, "ao-volume", MPV_FORMAT_INT64)) < 0)
        goto error;

    if ((status = mpv_observe_property(ctx, 0, "duration", MPV_FORMAT_INT64)) < 0)
        goto error;

    if ((status = mpv_observe_property(ctx, 0, "time-pos", MPV_FORMAT_INT64)) < 0)
        goto error;

    if ((status = mpv_observe_property(ctx, 0, "pause", MPV_FORMAT_FLAG)) < 0)
        goto error;

    fprintf(stderr, " * lmpv_new: successfully created a mpv context\n");
    return ctx;
error:
    lmpv_free(&ctx);
    return NULL;
}

/*
 *
 * NOW: THE SETTED FUNCTIONS FOR THE WINDOW MODULE :NOW
 *
 */

bool
module_x11_new(struct lkt_win *const win)
{
    (void) lx11_fullscreen;

    if (win == NULL)
        return false;

    struct module_x11_window *x11_win = win->window;

    if (x11_win == NULL) {
        x11_win = calloc(1, sizeof(struct module_x11_window));
        memset(x11_win, 0, sizeof(struct module_x11_window));

        if (x11_win == NULL) {
            fprintf(stderr, " ! module_x11_window: failed to allocate window, no memory\n");
            return false;
        }

        if (!lx11_new(x11_win)) {
            fprintf(stderr, " ! module_x11_window: failed to create the X11 window\n");
            return false;
        }

        fprintf(stderr, " . module_x11_window: successfully created the X11 window\n");
    }

    if (x11_win->mpv == NULL) {
        x11_win->mpv = lmpv_new(x11_win->master_win);
        if (x11_win->mpv == NULL)
            return false;
    }

    win->window = x11_win;
    return true;
}

void
module_x11_close(struct lkt_win *const win)
{
    if (win == NULL || win->window == NULL)
        return;

    struct module_x11_window *const x11_win = win->window;
    lmpv_free(&x11_win->mpv);
    x11_win->child_display = NULL;
    x11_win->child_win = 0;
}

void
module_x11_free(struct lkt_win *const win)
{
    if (win == NULL || win->window == NULL)
        return;

    struct module_x11_window *const x11_win = win->window;
    lmpv_free(&x11_win->mpv);

    if (win == NULL)
        return;

    XCloseDisplay(x11_win->master_display);
    free(x11_win);

    win->window = NULL;
}

bool
module_x11_toggle_pause(struct lkt_win *const win, bool *new_paused)
{
    if (win == NULL || win->window == NULL)
        return false;

    mpv_handle *ctx = ((struct module_x11_window *) win->window)->mpv;

    if (!ctx) {
        fprintf(stderr, " ! module_x11_toggle_pause: failed due to missing mpv ctx\n");
        return false;
    }

    const char *cmd[] = {"cycle", "pause", "up", NULL};
    int status;

    if ((status = mpv_command(ctx, cmd)) < 0) {
        fprintf(stderr, "mpv_toggle_pause: Failed issue command: %s\n",
                mpv_error_string(status));
        return false;
    }

    return module_x11_is_paused(win, new_paused);
}

bool
module_x11_load_file(struct lkt_win *const win, const char *filepath)
{
    if (win == NULL || win->window == NULL)
        return false;

    struct module_x11_window *xwin = win->window;
    mpv_handle *ctx = xwin->mpv;

    if (!ctx) {
        fprintf(stderr, " ! module_x11_load_file: failed due to missing mpv ctx\n");
        return false;
    }

    if (access(filepath, R_OK)) {
        fprintf(stderr, " ! mpv_add_file: Faild to read file: %s\n", filepath);
        return false;
    }

    const char *cmd[] = {"loadfile", filepath, "replace", NULL};
    int status;

    if ((status = mpv_command(ctx, cmd)) < 0) {
        fprintf(stderr, " ! mpv_add_file: Failed to add '%s': %s\n",
                filepath, mpv_error_string(status));
        return false;
    }

    xwin->mpv_duration = 0;
    xwin->mpv_time_pos = 0;

    fprintf(stderr, " . module_x11_load_file: added file %s\n", filepath);
    return true;
}

bool
module_x11_set_volume(struct lkt_win *const win, int vol)
{
    if (win == NULL || win->window == NULL)
        return false;

    mpv_handle *ctx = ((struct module_x11_window *) win->window)->mpv;

    if (!ctx) {
        fprintf(stderr, " ! module_x11_set_volume: failed due to missing mpv ctx\n");
        return false;
    }

    int status;
    char str[5];
    memset(str, 0, 5);
    snprintf(str, 4, "%d", vol);
    const char *cmd[] = {"set", "ao-volume", str, NULL};

    if ((status = mpv_command(ctx, cmd)) < 0) {
        fprintf(stderr, " ! module_x11_set_volume: Failed to execute command: %s\n",
                mpv_error_string(status));
        return false;
    }

    return true;
}

bool
module_x11_is_paused(struct lkt_win *const win, bool *ret)
{
    if (win == NULL || win->window == NULL)
        return false;

    mpv_handle *ctx = ((struct module_x11_window *) win->window)->mpv;
    int status;

    if (!ctx) {
        fprintf(stderr, " ! module_x11_is_paused: failed due to missing mpv ctx\n");
        return false;
    }

    if ((status = mpv_get_property(ctx, "pause", MPV_FORMAT_FLAG, &ret)) < 0) {
        fprintf(stderr, "mpv_is_paused: Failed to get pause property: %s\n",
                mpv_error_string(status));
        return false;
    }

    return true;
}

bool
module_x11_get_duration(struct lkt_win *const win, int *dur_sec)
{
    if (win == NULL || win->window == NULL)
        return false;

    struct module_x11_window *win_x11 = win->window;
    *dur_sec = win_x11->mpv_duration;

    return true;
}

bool
module_x11_get_elapsed(struct lkt_win *const win, int *elapsed_sec)
{
    if (win == NULL || win->window == NULL)
        return false;

    struct module_x11_window *win_x11 = win->window;
    *elapsed_sec = win_x11->mpv_time_pos;

    return true;
}

bool
module_x11_handle_events(struct lkt_win *const win, sqlite3 *db, enum mpd_idle_flag *mpd_idle_events)
{
    if (win == NULL || win->window == NULL)
        return false;

    bool sta = true;
    sta &= lx11_handle(win->window);
    sta &= lmpv_handle(win, db, mpd_idle_events);
    return sta;
}
